<?php
//====================================
//  rt.buffer HTTP handler
//------------------------------------
//  Scannex Electronics Limited
//  http://www.scannex.com
//  Copyright (c) UK 2018-2020
//  All Rights Reserved Worldwide
//------------------------------------
// 2019-03-06: MP: Simplified PHP
// 2020-05-05: MP: Started modifications for Abyss PHP
// 2020-05-06: MP: Bug fixes and additions (ref: AMJ)
// 2020-05-18: MP: Support for "results.txt" in update.txt processing
// 2020-06-08: MP: IIS compatibility fixes
// 2021-06-21: MP: Added support for decompressing .gz files for rt.buffer firmware 3.43+
//====================================

//------------------------------------
// User parameters for the PHP script
//------------------------------------
// PLEASE SET THE VALUES BELOW!
//-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v-v

$ROOT_DIRECTORY = 'c:/rtb/rt.buffer';

// The following value apply:
//  {root} = Root directory
//  {time} = Time
//  {name} = Buffer name
//  {serial} = Serial
//  {filename} = filename (as sent by rt.buffer)
//  {fileext} = extension of the filename (as sent by rt.buffer)
//  {date} = Date Stamp

$FILE_DATA      = '{root}/Data/{name}/{filename}';                        // Filename for incoming data
$FILE_DIAG      = '{root}/Info/{name}/{name}-{serial}-{time}-diag.txt';   // Filename for diagnostic dumps
$FILE_CELL      = '{root}/Info/{name}/{name}-{serial}-{time}-cell.txt';   // Filename for cellular info
$DIR_UPDATE     = '{root}/Updates/{name}/';                               // Directory for the updates
$UPDATE_FILENAME = 'update.txt';
$ERRORS_LOG      = '{root}/rtbuffer.php.errors.log';

$FILE_CHMOD      = 0777;               // When creating new directories for the buffer
$TIMEZONE        = 'UTC';              // For reporting back UTC time changes
$MAXTIMEDIFF     = 30;

$SHOW_WEB_OUTPUT = false;             // Display the web output for all clients (!! SECURITY DANGER !!)

$REQUIRE_SECRET = false; //true;  // Enable this for the HMAC-SHA256 signature exchange, and ideally fill in SECRET_HASH/HASHES (not SECRET)
//$SECRET = 'Secret!';
//$SECRET_HASH = 'e6b01082296dedaa16c406e5f634f3e906942839d2cb50aa5b64ecced7adcd49';  // |217.37.194.108|
// An array that allows multiple host names to be served by this one PHP script
//$SECRET_HASHES = array();
//$SECRET_HASHES['217.37.194.108'] = 'e6b01082296dedaa16c406e5f634f3e906942839d2cb50aa5b64ecced7adcd49'; // |217.37.194.108|
//...............................................................................
// NOTE: $SECRET_HASH + $SECRET_HASHES values are calculated as follows (in PHP):
//       = hash_hmac('sha256', '|'.$host_id.'|', 'RTB1:'.$SECRET);
//
// e.g. With the rt.buffer set with /cfg.key.set('iot', 'Secret!'), the PHP side is as follows:
//       = hash_hmac('sha256', '|217.37.194.108|RTB1:Secret!')
//       = 'e6b01082296dedaa16c406e5f634f3e906942839d2cb50aa5b64ecced7adcd49'
//...............................................................................

//PHP Error Logging
	ini_set('error_log', $ERRORS_LOG);
	ini_set('log_errors', 1);
	ini_set('error_reporting', E_ALL & ~E_NOTICE);

//-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^
// PLEASE SET THE VALUES ABOVE!
//------------------------------------

//------------------------------------
// Global parameters for the script
// DO NOT MODIFY!
//------------------------------------
$RTBUFFER_UPLOAD_DATE = '2020-05-14';
$RTBUFFER_UPLOAD_VERSION = '1.11';

// Strings for the manipulation of the file names and directories
$findstrings = array('{root}', '{time}', '{name}', '{serial}', '{filename}', '{fileext}', '{date}', '{yy}', '{mm}', '{dd}');
$replacestrings = array();
date_default_timezone_set($TIMEZONE); // Set the timezone
ini_set('auto_detect_line_endings', true);

function getServerValue($v)
{
  if (array_key_exists($v, $_SERVER))
    return @$_SERVER[$v];
  else
    return "";
}

//------------------------------------------------------
// Get IP Address of client
//------------------------------------------------------
function getRealIpAddr()
{
  if (!empty($_SERVER['HTTP_CLIENT_IP']))   //check ip from share internet
    $ip=$_SERVER['HTTP_CLIENT_IP'];
  elseif (!empty($_SERVER['HTTP_X_FORWARDED_FOR']))   //to check ip is pass from proxy
    $ip=$_SERVER['HTTP_X_FORWARDED_FOR'];
  else
    $ip=$_SERVER['REMOTE_ADDR'];
  return $ip;
}

//------------------------------------------------------
// Get IP Address of server
//------------------------------------------------------
function getServerAddress() {
  if(array_key_exists('SERVER_ADDR', $_SERVER))
    return $_SERVER['SERVER_ADDR'];
  elseif(array_key_exists('LOCAL_ADDR', $_SERVER))
    return $_SERVER['LOCAL_ADDR'];
  elseif(array_key_exists('SERVER_NAME', $_SERVER))
    return gethostbyname($_SERVER['SERVER_NAME']);
  else {
    // Running CLI
    if(stristr(PHP_OS, 'WIN')) {
        return gethostbyname(php_uname("n"));
    } else {
        $ifconfig = shell_exec('/sbin/ifconfig eth0');
        preg_match('/addr:([\d\.]+)/', $ifconfig, $match);
        return $match[1];
    }
  }
}

//------------------------------------------------------
// Return a date string suitable for a filename
//------------------------------------------------------
function DateStringFile($fmt = 'Ymd') {
  return date($fmt);
}

//------------------------------------------------------
// Return a time string suitable for a filename
//------------------------------------------------------
function TimeStringFile() {
  return date('Ymd-His');
}

//------------------------------------------------------
// Log error message
//------------------------------------------------------
function LogToFile($errcode, $msg='', $extra='') {
  global $ROOT_DIRECTORY, $ERRORS_LOG;
  $txt = TimeStringFile().",".getRealIpAddr().",".@$_POST['serial'].",".$errcode.",".$msg.",".$extra . "\r\n";
  error_log($txt, 3, ConvertName($ERRORS_LOG));
}

//------------------------------------------------------
// Abort with a helpful error message
//------------------------------------------------------
function FatalError($msg, $err=501, $extra='') {
  if ($err==501) $msg = 'ERROR: '.$msg;
  LogToFile($err, $msg, $extra);
  header('HTTP/1.1 '.$err.' '.$msg);
  ob_end_flush();
  die($msg);
  exit(1);
}

//------------------------------------------------------
// Set up the variables for translation into ConvertName
//------------------------------------------------------
function SetupVars($name, $serial) {
  global $replacestrings, $ROOT_DIRECTORY;
  $replacestrings = array($ROOT_DIRECTORY    //{root}
                         ,TimeStringFile()   //{time}
                         ,$name              //{name}
                         ,$serial            //{serial}
                         ,''                 //{filename}
                         ,''                 //{fileext}
                         ,DateStringFile()   //{date}
                         ,DateStringFile('Y')//{yy}
                         ,DateStringFile('m')//{mm}
                         ,DateStringFile('d')//{dd}
                         );

}

//------------------------------------------------------
// Set up the filename variable for ConvertName
//------------------------------------------------------
function SetupVarsFilename($filename, $check=true) {
  global $replacestrings;
  if ($check)
  {
    $filename = basename($filename, '.zlib');
    $filename = basename($filename, '.gz'); // Support GZ (see also gzopen)
    CheckValue($filename);
  }
  $replacestrings[4] = $filename;
  $fileext = pathinfo($filename, PATHINFO_EXTENSION);
  if (!$fileext) $fileext = 'dat';
  $replacestrings[5] = $fileext;
}

//------------------------------------------------------
// Convert a template into a real filename
//------------------------------------------------------
function ConvertName($fmt) {
  global $findstrings;
  global $replacestrings;
  return str_replace($findstrings, $replacestrings, $fmt);
}

//------------------------------------------------------
// Returns true if $s ends with the string $pattern
//------------------------------------------------------
function strendswith($s, $pattern, $case=true) {
  $s = substr($s, -strlen($pattern));
  if (!$case) $s = strtolower($s);
  return ($s == $pattern);
}

//------------------------------------------------------
// Find a file and return just filename
//------------------------------------------------------
function FindAFile($dir, $pattern) {
  $ar = glob($dir.$pattern);
  foreach ($ar as $x) {
    if (!strpos($x, '.error')) {
      return basename($x);
    }
  }
  return false;
}

//------------------------------------------------------
// Find a file and return full path+filename
//------------------------------------------------------
function FindAFullFile($dir, $pattern) {
  $dir = ConvertName($dir);
  $fn = FindAFile($dir, $pattern);
  if ($fn) $fn = $dir.$fn;
  return $fn;
}

//------------------------------------------------------
// Create directory if needed
//------------------------------------------------------
function EnsureDirectory($dir) {
  global $FILE_CHMOD;
  $dir = ConvertName($dir);
  @mkdir($dir, $FILE_CHMOD, true);
}

//------------------------------------------------------
// Copy a file called $filename from $sourcename to $target
//------------------------------------------------------
function CopyFile($sourcename, $filename, $target) {
  $zlib = false;
  /*
  echo $sourcename . "\r\n";
  echo $filename . "\r\n";
  echo $target . "\r\n";
  */

  EnsureDirectory(dirname($target));
  $src = gzopen($sourcename, 'rb');  // Allow GZIP detection
  if (!$src) { FatalError('Error opening source file'); }
  $dst = fopen($target, 'wb');
  if (!$dst) { FatalError('Error opening target file'); }
  if (isset($filename) && strendswith($filename, '.zlib', false))
  {
    // Check that first byte & 0x0f == 0x08
    // Check that (byte[0] * 256 + byte[1]) % 31 = 0
    $zlibheader = fread($src, 2);
    if ( ((ord($zlibheader[0]) & 0x0f) != 8) ||
         ( (((ord($zlibheader[0]) * 256) + ord($zlibheader[1])) % 31) != 0)
        )
      FatalError('Incorrect zlib header!');
    fseek($src, 2, SEEK_SET);
    stream_filter_append($src, 'zlib.inflate', STREAM_FILTER_READ);
  }
  while (false != ($txt = fread($src, 4096))) {
    if (strlen($txt) == 0) break;
    fwrite($dst, $txt);
  }
  fclose($dst);
  fclose($src);
}

//------------------------------------------------------
// Validate a string for 'nasty' characters
//------------------------------------------------------
function CheckValue($str, $blankna=true) {
  if (isset($str)) {
    $illegalchars = "\\/\"':";
    if (strpbrk($str, $illegalchars)) {
      FatalError('Invalid request', 400, $str);
    }
  }
  if ($blankna && ($str == "")) {
    FatalError('Invalid Request', 400);
  }
}

//------------------------------------------------------
// Pull a parameter from either the header or the POST body
//------------------------------------------------------
function GetParam($name) {
  $result = @getServerValue('HTTP_RTBUFFER_'.strtoupper($name));
  if (!$result) $result = @$_POST[$name];
  return $result;
}

//------------------------------------------------------
// Separate the HTTP parameters
// Used for 'rtbuffer-req: verb; field=value; field=value'
// The returned array has the verb in $results[0]
// All other values are $results['field']=value
//------------------------------------------------------
function ExtractParameters($value) {
  $results = [];
  $parts = explode(';', $value);
  foreach ($parts as $i => $v) {
    $part = explode('=', $v);
    if ($i==0) {
      $results[] = [trim(@$part[0]), trim(@$part[1])];
    } else {
      $results[trim(@$part[0])] = trim(@$part[1]);
    }
    //$parts[$i] = trim($v);
  }
  return $results;
}

//------------------------------------------------------
// (DEBUG) Convert array to text
//------------------------------------------------------
function ParamArrayToText($pa) {
  $result = '';
  foreach ($pa as $p => $v) {
    $result = $result .'|'. $p.'='.$v;
  }
  return $result;
}

//------------------------------------------------------
// Append a header and add the value to the hash list
//------------------------------------------------------
function ReplyStart() {
  global $HASH_DATA;
  $HASH_DATA = getServerValue('HTTP_RTBUFFER_SHA');
}

//------------------------------------------------------
// Append a header and add the value to the hash list
//------------------------------------------------------
function ReplyHeader($key, $value) {
  global $HASH_DATA;
  $value = trim($value);
  header($key.': '.$value);
  $HASH_DATA = $HASH_DATA . '|'. $value;
}

//------------------------------------------------------
// Append the 'rtbuffer-ack' value
//------------------------------------------------------
function ReplyAck($value) {
  ReplyHeader('rtbuffer-ack', $value);
}

//------------------------------------------------------
// Calculate the HMAC-SHA256 and add to the reply
//------------------------------------------------------
function ReplySha($hashoffile = null) {
  global $HASH_DATA, $SECRET_HASH, $REQUIRE_SECRET;
  if (!$REQUIRE_SECRET) return;
  if (!$hashoffile) $hashoffile = hash('sha256', ''); // Empty
  $h2 = hash_hmac('sha256', $HASH_DATA .'|'. $hashoffile, $SECRET_HASH);
  //!!//header('rtbuffer-hasher: '.$HASH_DATA); //!!//
  header('rtbuffer-sha: '.$h2);
}

//------------------------------------------------------
// Send back the file
//------------------------------------------------------
function ReplyFileContents($file) {
  global $HASH_DATA;
  ReplyAck('1');  // Stuff to do
  header('Content-Type: application/octet-stream');
  header('Content-Length: '. filesize($file));
  $hf = hash_file('sha256', $file);  // Calculate the hash of the file
  //!!//header('rtbuffer-hf: '.$hf); //!!//
  ReplySha($hf);
  ob_end_flush();
  @readfile($file);
}

//------------------------------------------------------
// Send back the update.txt file (if exists)
//------------------------------------------------------
function ProcessUpd($serial, $name, $updatefile, $req) {
  // Hand back the 'update.txt' file... if exists
  if (!$updatefile) {
    ReplyAck('0');  // NOTHING to do
    ReplySha();
    return;
  }
  ReplyFileContents($updatefile);
}

//------------------------------------------------------
// Handle the process compelete (ack or nak)
//------------------------------------------------------
function ProcessUpdateComplete($serial, $name, $updatefile, $req, $ok) {
  $pp = pathinfo($updatefile);
  $newname = $pp['dirname'] .'/'. $pp['filename'] .'.'. TimeStringFile() .'.'. $pp['extension'];
  rename($updatefile, $newname);
  if (!$ok) {
    $msg = $req['msg'];
    if ($msg) {
      $haveresults = false;
      // Save the error message in a side-car file
      // Save the contents of "results.txt" if present...
      foreach (@$_FILES as $key => $val) {
        $filename = $val['name'];
        SetupVarsFilename($filename);
        if ($key == 'results') {
          if ($val['error'] == 0) {
            CopyFile($val['tmp_name'], $filename, $newname.'.error.txt');
            $haveresults = true;
          }
        }
      }
      if (!$haveresults) file_put_contents($newname.'.error.txt', $msg, LOCK_EX);
    }
  }
  ReplyAck('1');  // Tell the buffer we've done this ok
  ReplySha();
}
function ProcessAck($serial, $name, $updatefile, $req) {
  return ProcessUpdateComplete($serial, $name, $updatefile, $req, true);
}
function ProcessNak($serial, $name, $updatefile, $req) {
  return ProcessUpdateComplete($serial, $name, $updatefile, $req, false);
}

//------------------------------------------------------
// Convert relative to absolute
//------------------------------------------------------
function get_absolute_path($path) {
  $path = str_replace(array('/', '\\'), DIRECTORY_SEPARATOR, $path);
  $parts = array_filter(explode(DIRECTORY_SEPARATOR, $path), 'strlen');
  $absolutes = array();
  foreach ($parts as $part) {
      if ('.' == $part) continue;
      if ('..' == $part) {
          array_pop($absolutes);
      } else {
          $absolutes[] = $part;
      }
  }
  return implode(DIRECTORY_SEPARATOR, $absolutes);
}

//------------------------------------------------------
// Check whether $prefix is at the beginning of $path
// Should work on Windows + Unix
//------------------------------------------------------
function IsPathPrefix($full, $prefix)
{
  $full = str_replace("\\", "/", $full);
  $prefix = str_replace("\\", "/", $prefix);
  return strpos($full, $prefix) === 0;
}

//------------------------------------------------------
// Convert a filename for this buffer
//------------------------------------------------------
function ConvertRelativeFilename($fn) {
  global $DIR_UPDATE, $ROOT_DIRECTORY;
  // Convert to a real path...
  if (($fn[0] == '/') || ($fn[0] == '\\')) {
    $path = $ROOT_DIRECTORY.$fn;          // Pseudo root
  } else {
    $path = ConvertName($DIR_UPDATE).$fn; // Relative to this buffer
  }
  $path = get_absolute_path($path);

  // Check that the path is still a child of the ROOT_DIRECTORY...
  if (!IsPathPrefix($path, $ROOT_DIRECTORY)) {
    // Security violation!
    FatalError('Invalid filename.');
    return;
  }
  return $path;
}

//------------------------------------------------------
// Remove the 'path' that has tilde characters
//------------------------------------------------------
function RemoveTilde($str) {
  $p = strrchr($str, '~');
  if ($p === false) {
    // Not found
    return $str;
  } else {
    return substr($p, 1);  // Skip the '~'
  }
}

//------------------------------------------------------
// Save any attached file(s)
//------------------------------------------------------
function ProcessPut($serial, $name, $updatefile, $req) {
  global $FILE_DATA;
  global $FILE_CELL;
  global $FILE_DIAG;

  if ($updatefile && @$_FILES['data']) {
    // This is a data push, but tell the buffer there's something to do
    ReplyHeader('rtbuffer-upd', 'now');
  }

  foreach (@$_FILES as $key => $val) {
    $filename = $val['name'];
    if ($key != 'file') {
      $filename = RemoveTilde($filename);   // Just pure filename
    }
    SetupVarsFilename($filename);
    if ($key == 'data') {
      $target = ConvertName($FILE_DATA);    // Data files
    } elseif ($key == 'cellinfo') {
      $target = ConvertName($FILE_CELL);    // Cellular info
    } elseif ($key == 'diag') {
      $target = ConvertName($FILE_DIAG);    // Diagnostic dump
    } elseif ($key == 'file') {
      // Use the general filename, not the Multipart name...
      $target = ConvertRelativeFilename($req['fn']); // Special case for diag pushed files - work out relative path
    } else {
      FatalError('Unknown file type');
    }
    if ($val['error'] == 0) {
      CopyFile($val['tmp_name'], $filename, $target);
    } else {
      var_dump($val); //????
      FatalError('Error with file '.$filename);
    }
  }
  ReplyAck('1');
  ReplySha();
}

//------------------------------------------------------
// Send a file to the rt.buffer
//------------------------------------------------------
function ProcessGet($serial, $name, $updatefile, $req) {
  $fn = @$req['fn'];
  if (!$fn) {
    FatalError('Invalid filename '.$fn);
    return;
  }
  $path = ConvertRelativeFilename($fn);
  // Check whether the file is available
  if (!file_exists($path)) {
    // File does not exist!
    ReplyAck('0');
    ReplySha();
    return;
  }
  ReplyFileContents($path);
}

//------------------------------------------------------
// Verify the HMAC-SHA256 from the rt.buffer request
//------------------------------------------------------
function VerifyRequest() {
  global $SECRET, $SECRET_HASH, $SECRET_HASHES, $REQUIRE_SECRET;

  if (!$REQUIRE_SECRET) return true;
  // Ideally you would only use $SECRET_HASH so that $SECRET cannot be extracted
  // But the hash is only valid for this host + URI combination. i.e. you cannot deploy elsewhere
  // NOTE: Does not work for mixed 'name' vs 'dotted' host!!! Use one or the other
  $host_id = strtok($_SERVER['HTTP_HOST'], ':');
  $secret_hash = $SECRET_HASHES[$host_id];
  if ($secret_hash) $SECRET_HASH = $secret_hash;

  if (!$SECRET_HASH || ($SECRET_HASH==''))
  {
    // Need to calculate the secret_hash
    //!!//header("rtbuffer-hasher1: ".$data); //!!//
    $SECRET_HASH = hash_hmac('sha256', '|'.$host_id.'|', 'RTB1:'.$SECRET);
    //!!//header('rtbuffer-dbghmac1: '.$SECRET_HASH); //!!// can be saved in the PHP instead of the secret
  }

  $data = trim(getServerValue('REQUEST_METHOD')) .'|'.
          trim(getServerValue('HTTP_RTBUFFER_REQ')) .'|'.
          trim(getServerValue('HTTP_RTBUFFER_SERIAL')) .'|'.
          trim(getServerValue('HTTP_RTBUFFER_NAME')) .'|'.
          trim(getServerValue('HTTP_RTBUFFER_UTC'));

  // Standard hash the contents of all the names, filenames, and contents...
  //!!!!//var_dump($_FILES);  //!!!!//
  $ctx = hash_init('sha256');
  foreach (@$_FILES as $key => $val) {
    // NOTE: These MUST iterate in the order they were posted!!!
    //!!//header('rtbuffer-file-'.$key.': |'.$val['name'].'|'); //!!//
    hash_update($ctx, $key.'|'.$val['name'].'|'); // Type + Filename
    hash_update_file($ctx, $val['tmp_name']);     // Contents (from the HTTP attachment)
    hash_update($ctx, '|');
  }
  // Need to finalise the hash, and add the ASCII Hex to the HMAC block
  $sha = hash_final($ctx);
  //!!//header('rtbuffer-file: '.$sha); //!!//
  $data = $data .'|'. $sha;

  //!!//header('rtbuffer-hasher2: '.$data); //!!//
  $h2 = hash_hmac('sha256', $data, $SECRET_HASH);
  //!!//header('rtbuffer-dbghmac2: '.$h2); //!!//
  return ($h2 == getServerValue('HTTP_RTBUFFER_SHA'));
}

//------------------------------------------------------
// Handle the POST request
//------------------------------------------------------
function ProcessPostRequest() {
  global $FILE_DATA, $FILE_DIAG, $FILE_CELL;
  global $MAXTIMEDIFF;
  global $DIR_UPDATE, $UPDATE_FILENAME;

  $serial = GetParam('serial');
  $name   = GetParam('name');

  if (!$serial) FatalError('Incorrect request', 400); // we must at least have a serial number!
  CheckValue($serial);        // Verify that serial number hasn't been cooked by a malicious client

  if (($name == "") || ($name == "Scannex"))
  {
    $name = 'Scannex-'.$serial; // Use the serial number if the rt.buffer name hasn't been programmed
  }
  CheckValue($name, false);   // Don't allow directory or drive characters in the name or serial
  SetupVars($name, $serial);  // Get ready for the Replace
  ReplyStart();

  // Check if time needs updating
  $utc = getServerValue('HTTP_RTBUFFER_UTC');
  if ($utc) {
    if (abs(intval($utc) - time()) > $MAXTIMEDIFF)
      ReplyHeader('rtbuffer-utc', ''.time());
  }

  EnsureDirectory($DIR_UPDATE);
  $updatefile = FindAFullFile($DIR_UPDATE, $UPDATE_FILENAME);

  if (!VerifyRequest()) {
    //DEBUG//header('rtbuffer-ERROR: Unauthorized'); // DEBUG: test notification instead of fatal_error
    FatalError('Unauthorized secret', 401, 'Unauthorized');
    exit;
  }

  // Pull out additional parameters and convert to array
  $req = getServerValue('HTTP_RTBUFFER_REQ'); // Verb for the server "upd", "ack", "nak", "put", "get"
  $reqp = ExtractParameters($req);      // Rip apart any additional parameters
  list($req) = array_shift($reqp);      // Pull off the first value as the new $req

  // Vector to the specific handlers...
  if     ($req == 'upd') { ProcessUpd($serial, $name, $updatefile, $reqp); }
  elseif ($req == 'ack') { ProcessAck($serial, $name, $updatefile, $reqp); }
  elseif ($req == 'nak') { ProcessNak($serial, $name, $updatefile, $reqp); }
  elseif ($req == 'put') { ProcessPut($serial, $name, $updatefile, $reqp); }
  elseif ($req == 'get') { ProcessGet($serial, $name, $updatefile, $reqp); }
  else                   { header('rtbuffer-err: Unknown request'); }
}

//------------------------------------------------------
// Give a localhost status page
//------------------------------------------------------
function ProduceWebOutput() {
  global $RTBUFFER_UPLOAD_VERSION, $RTBUFFER_UPLOAD_DATE, $SHOW_WEB_OUTPUT;
  global $ROOT_DIRECTORY, $FILE_DATA, $FILE_DIAG, $FILE_CELL, $DIR_UPDATE;
  global $replacestrings, $findstrings;

  $ServerAddress = getServerAddress();

  header("Content-type: text/html");
  echo "<html>";
  echo "<style>th{text-align:right;width:200px;}</style>";
  echo "<body><font face='Verdana, Trebuchet MS, Arial'>";
  echo "<h1>Cloud Server for rt.buffer</h1>";
  echo "(c) UK 2019-2020 Scannex Electronics Limited. All rights reserved worldwide.</br>";
  echo "<a href='http://www.scannex.com'>www.scannex.com</a>";
  echo "<hr>";
  if (getServerValue("HTTPS") && (getServerValue("HTTPS") != "off")) echo "<b>Secure Connection!</b>";

  // Check whether this is a local request, and serve up some helpful info...
  if ($SHOW_WEB_OUTPUT ||
    (($ServerAddress == "127.0.0.1") && ($_SERVER["REMOTE_ADDR"] == "127.0.0.1")) ||
    (($_SERVER["SERVER_ADDR"] == "127.0.0.1") && ($_SERVER["REMOTE_ADDR"] == "127.0.0.1")) ||
    (($_SERVER["SERVER_ADDR"] == "::1") && ($_SERVER["REMOTE_ADDR"] == "::1"))) {
    // Script information about this server
    SetupVars(
            "<font color='red'><em>name</em></font>",
            "<font color='red'><em>serial</em></font>");
    SetupVarsFilename(
            "<font color='red'><em>filename</em></font>", false
           );

    echo "<h2>Application Info</h2>";
    echo "<table border=1 cellpadding=5>";
    echo "<tr><th>Version</th><td>".$RTBUFFER_UPLOAD_VERSION." (".$RTBUFFER_UPLOAD_DATE.")</td></tr>";
    echo "<tr><th>Host</th><td>".$_SERVER["HTTP_HOST"]." (".$ServerAddress.")</td></tr>";
    echo "<tr><th>Remote</th><td>".getRealIpAddr()."</td></tr>";
    echo "<tr><th>URL Path</th></td><td>".$_SERVER["REQUEST_URI"]."</td></tr>";
    echo "<tr><th>Local Path</th></td><td>".$_SERVER["SCRIPT_FILENAME"]."</td></tr>";
    echo "</table>";

    echo "<h2>Directories</h2>";
    echo "<table border=1 cellpadding=5>";
    echo "<tr><th>Root Directory</th></td><td>".$ROOT_DIRECTORY."</td></tr>";
    echo "<tr><th>Data File</th></td><td>".ConvertName($FILE_DATA)."</td></tr>";
    echo "<tr><th>Diagnostic File</th></td><td>".ConvertName($FILE_DIAG)."</td></tr>";
    echo "<tr><th>Cellular File</th></td><td>".ConvertName($FILE_CELL)."</td></tr>";
    echo "<tr><th>Update Directory</th></td><td>".ConvertName($DIR_UPDATE)."</td></tr>";
    echo "</table>";

    echo "<h2>Tags</h2>";
    echo "<table border=1 cellpadding=5>";
    foreach($findstrings as $index=>$tag) {
      echo "<tr><th>".$tag."</th></td><td>".@$replacestrings[$index]."</td></tr>";
    }
    echo "</table>";
    //var_dump($replacestrings);

  }
  echo "</body></html>";
}

//------------------------------------
// MAIN start
//------------------------------------
try {
  ob_start();   // Start buffering so we can adjust the headers as needed
  if ($_SERVER['REQUEST_METHOD'] == 'POST')
    ProcessPostRequest(); // Handle the post
  else
    ProduceWebOutput();   // GET and everything else. Let's display some helpful information
  @ob_end_flush();  //End the buffering (may have been ended already by the update process)
}
catch(Exception $e) {
  FatalError('Exception: '.$e->getMessage());
}
?>